from typing import List, Tuple
from sqlalchemy import Table, Column, MetaData
from components.Databases.Mysql import Mysql
from components.Databases.Mariadb import Mariadb
from components.Databases.Postgres import Postgres
from components.Databases.Oracle import Oracle

from enums.DatabaseTypeEnum import DatabaseTypeEnum
import logging
import sys
import oracledb

oracledb.version = "8.3.0"
sys.modules["cx_Oracle"] = oracledb
logger = logging.getLogger()
metadata_obj = MetaData()


class Database:

    def __init__(self, db_type: str, host: str, port: int, user: str, password: str, database: str,
                 service: str, schema: str) -> None:
        self.db_type = db_type
        if db_type == DatabaseTypeEnum.MYSQL.value:
            self.factory = Mysql(host, port, user, password, database)
        elif db_type == DatabaseTypeEnum.MARIADB.value:
            self.factory = Mariadb(host, port, user, password, database)
        elif db_type == DatabaseTypeEnum.POSTGRES.value:
            self.factory = Postgres(host, port, user, password, database, schema)
        elif db_type == DatabaseTypeEnum.ORACLE.value:
            self.factory = Oracle(host, port, user, password, service)
        self.engine = None
        self.procedure_identifier = "execute"

    def get_basic_connection(self):
        connection = None
        try:
            connection = self.factory.get_basic_connection()
        except Exception as e:
            logger.error(f"Error trayendo básica conexión. {e}")
        finally:
            return connection

    def close_basic_connection(self) -> None:
        try:
            self.factory.connection.close()
            self.factory.connection = None
        except Exception as e:
            logger.error(f"Error cerrando básica conexión. {e}")

    def create_engine(self) -> None:
        try:
            if isinstance(self.engine, type(None)):
                self.factory.create_engine()
                self.engine = self.factory.engine
        except Exception as e:
            logger.error(f"Error creando db engine. {e}")

    def create_model(self, tablename: str, fields: List[Tuple[str]], db_target: str,
                     modelName: str = "TableModel") -> bool:
        create = False
        try:
            create = self.factory.create_model(tablename, fields, db_target, modelName)
        except Exception as e:
            raise AssertionError(f"Error creando tabla dinámica con nombre {tablename}. {e}")
        finally:
            return create

    def create_table(self, model) -> bool:
        save = False
        try:
            model.metadata.create_all(self.engine)
            save = True
        except Exception as e:
            logger.error(f"Error creando tabla {model}. {e}")
        finally:
            return save

    def generate_sql_procedure(self, command: str) -> str:
        response = ""
        try:
            command = command.lower()
            response = self.factory.generate_sql_procedure(command, self.procedure_identifier)
        except Exception as e:
            logger.error(f"Error generando comando sql para procedure. Comando: {command}. {e}")
        finally:
            return response

    def get_all_tablenames(self) -> List[str]:
        tablenames = []
        try:
            tablenames = self.factory.get_all_tablenames()
        except Exception as e:
            logger.error(f"Error obteniendo los nombres de tablas. {e}")
        finally:
            return tablenames

    def check_procedure(self, procedure_name, connection) -> bool:
        result = False
        try:
            result = self.factory.check_procedure(procedure_name, connection)
        except Exception as e:
            logger.error(f"Error obteniendo los nombres de tablas. {e}")
        finally:
            return result

    def check_table(self, table_name, connection) -> bool:
        result = False
        try:
            result = self.factory.check_table(table_name, connection)
        except Exception as e:
            logger.error(f"Error obteniendo los nombres de tablas. {e}")
        finally:
            return result

    def verify_table(self, table_name, connection) -> bool:
        result = False
        try:
            result = self.factory.verify_table(table_name, connection)
        except Exception as e:
            logger.error(f"Error obteniendo numero de registros de la tabla. {e}")
        finally:
            return result