from typing import List, Tuple

import pymysql
from sqlalchemy import create_engine
from sqlalchemy.orm import Session
from enums.DatabaseDialectEnum import DatabaseDialectEnum
from enums.DatabaseTypeEnum import DatabaseTypeEnum
from components.Model.InsumoModel import InsumoModel, InsumoModelOracle
from components.Databases.Enums.MysqlDataTypeEnum import MysqlDataTypeEnum
from components.Databases.Enums.MysqlDataTypeORMEnum import MysqlDataTypeORMEnum


from sqlalchemy import Column
import logging

logger = logging.getLogger()


class Mysql:

    def __init__(self, host: str, port: int, user: str, password: str, database: str) -> None:
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.engine = None
        self.connection = None

    def get_basic_connection(self):
        try:
            if isinstance(self.connection, type(None)):
                self.connection = pymysql.connect(host=self.host, port=self.port, user=self.user,
                                                  password=self.password, db=self.database)
        except Exception as e:
            logger.error(f"Error obteniendo conexion básica de Mysql. {e}")
        finally:
            return self.connection

    def create_engine(self) -> None:
        try:
            dialect = DatabaseDialectEnum.MYSQL.value
            url = f"{dialect}://{self.user}:{self.password}@{self.host}:{str(self.port)}/{self.database}?charset=utf8mb4"
            self.engine = create_engine(url)
        except Exception as e:
            logger.error(f"Error creando engine de Mysql. {e}")

    def create_model(self, tablename: str, fields: List[Tuple[str]], db_target: str, modelName: str = "TableModel"):
        model = None
        try:
            db_args = {
                '__tablename__': tablename,
                '__table_args__': {'extend_existing': True}
            }
            if db_target == DatabaseTypeEnum.ORACLE.value:
                model = type(modelName, (InsumoModelOracle,), db_args)
            else:
                model = type(modelName, (InsumoModel,), db_args)
            for field in fields:
                logger.debug(f"Attribute: {field}")
                name = field[0]
                if field[2] != -1:
                    size = int(field[2] / 4)
                    try:
                        if not isinstance(field[3], type(None)) and field[3] > 0:
                            precision, scale = field[2], field[3]
                            if scale > precision:
                                precision = field[3]
                                scale = field[2]
                            data_type = MysqlDataTypeORMEnum[MysqlDataTypeEnum(field[1]).name].value(precision=precision, scale=scale)
                        else:
                            data_type = MysqlDataTypeORMEnum[MysqlDataTypeEnum(field[1]).name].value(size)
                    except TypeError:
                        data_type = MysqlDataTypeORMEnum[MysqlDataTypeEnum(field[1]).name].value()
                else:
                    data_type = MysqlDataTypeORMEnum[MysqlDataTypeEnum(field[1]).name].value()
                setattr(model, name, Column(data_type))
            model = model.__table__
        except Exception as e:
            logger.error(f"Error creando modelo dinámico en Mysql con nombre {tablename}. {e}")
        finally:
            return model

    def generate_sql_procedure(self, command: str, reserved_word: str = "execute") -> str:
        response = ""
        try:
            command = command.replace(reserved_word, "").replace(";", "")
            response = f"call {command};"
            logger.debug("COMANDO MYSQL:", response)
        except Exception as e:
            logger.error(f"Error generando comando sql para procedure Mysql. Comando: {command}. {e}")
        finally:
            return response

    def get_all_tablenames(self) -> List[str]:
        tablenames = []
        try:
            command = f"SELECT table_name FROM information_schema.tables WHERE table_schema='{self.database}'"
            with self.engine.connect() as conn:
                tablenames = conn.execute(command).all()
        except Exception as e:
            logger.error(f"Error obteniendo los nombres de tablas. {e}")
        finally:
            return tablenames

    def check_procedure(self, procedure_name, connection) -> bool:
        exists = False
        try:
            check_query = f"SHOW CREATE PROCEDURE {procedure_name}"
            result = connection.execute(check_query)
            exists = bool(result.fetchone())
        except Exception as e:
            logger.error(f"Error obteniendo existencia de procedure. {e}")
        finally:
            return exists

    def check_table(self, table_name, connection) -> bool:
        exists = False
        try:
            check_query = f"SHOW CREATE TABLE {table_name}"
            result = connection.execute(check_query)
            exists = bool(result.fetchone())
        except Exception as e:
            logger.error(f"Error obteniendo existencia de tabla. {e}")
        finally:
            return exists

    def verify_table(self, table_name, connection) -> bool:
        exists = False
        try:
            check_query = f"SELECT COUNT(*) FROM {table_name}"
            result = connection.execute(check_query)
            if result > 0:
                exists = True
        except Exception as e:
            logger.error(f"Error obteniendo counts de tabla. {e}")
        finally:
            return exists