import fnmatch
import datetime
from typing import Any, Dict, List, Tuple
import pytz
from io import BytesIO, StringIO
import pandas as pd
import re

from components.Utils import get_type_file
from enums.FileTypeEnum import FileTypeEnum
from enums.ScriptFileTypeEnum import ScriptFileTypeEnum
from enums.ProviderTypeEnum import ProviderTypeEnum
from components.Timezone import datetime_by_tzone

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.contrib.hooks.gcs_hook import GoogleCloudStorageHook
import logging

logger = logging.getLogger()


def get_df_from_s3(conn: str, bucket: str, key: str, delimiter: str, base_date: datetime.date,
                   interval: str) -> Dict[str, Any]:
    response = {'filename': "", 'df': None}
    dataframe = None
    try:
        s3_data = get_data_from_s3(conn, bucket, key, base_date, interval)
        logger.info(f"ARCHIVO EXTRAIDO: {s3_data}")
        if s3_data["filename"] == "":
            raise Exception(f"No se encontró archivo para el key: {key} y fecha base {base_date} en intervalo {interval}")
        response.update({'filename': s3_data["filename"]})
        file_type = get_type_file(s3_data["filename"])
        if file_type == FileTypeEnum.EXCEL:
            dataframe = pd.read_excel(s3_data["data"], engine="openpyxl", dtype='object')
        elif file_type == FileTypeEnum.OLD_EXCEL:
            dataframe = pd.read_excel(s3_data["data"], engine="xlrd", dtype='object')
        elif file_type == FileTypeEnum.TEXT or file_type == FileTypeEnum.CSV:
            str_data = str(s3_data["data"].getvalue(), encoding='UTF-8', errors='ignore')
            data = StringIO(str_data)
            dataframe = pd.read_csv(data, sep=delimiter, dtype='object')
        response.update({'df': dataframe})
    except Exception as e:
        logger.error(f"Error trayendo y transformando a DataFrame desde S3. {e}")
    return response


def get_data_from_s3(conn: str, bucket: str, key: str, base_date: datetime.date, interval: str) -> Dict[str, Any]:
    result = {'filename': '', 'data': BytesIO()}
    try:
        if key.rfind("/") != -1:
            prefix = key[:key.rfind("/")+1]
        else:
            prefix = ""
        s3_hook = S3Hook(conn)
        files = s3_hook.list_keys(bucket, prefix)
        last_key = ("", base_date)
        # Colocar una fecha muy atrás como base
        for file_key in files:
            if fnmatch.fnmatch(file_key, key):
                file_date = s3_hook.get_key(file_key, bucket).meta.data
                file_date = file_date["LastModified"]
                if int(interval) - abs((file_date - last_key[1]).days) >= 0:
                    last_key = (file_key, file_date)
        data = s3_hook.get_key(last_key[0], bucket)
        data.download_fileobj(result["data"])
        result["filename"] = last_key[0]
    except Exception as e:
        logger.error(f"Error trayendo datos desde S3 para el key {key}. {e}")
    return result


def get_base_date(conn: str, bucket: str, key: str) -> datetime.date:
    utc = pytz.UTC
    # Colocar una fecha muy atrás como base
    last_date = datetime.datetime(2000, 1, 1, 0, 0, 0).replace(tzinfo=utc)
    try:
        if key.rfind("/") != -1:
            prefix = key[:key.rfind("/") + 1]
        else:
            prefix = ""
        s3_hook = S3Hook(conn)
        files = s3_hook.list_keys(bucket, prefix)
        for file_key in files:
            if fnmatch.fnmatch(file_key, key):
                file_date = s3_hook.get_key(file_key, bucket).meta.data
                file_date = file_date["LastModified"]
                if last_date >= file_date:
                    continue
                last_date = file_date
        logger.debug(f"Fecha base desde {key} : {last_date}")
    except Exception as e:
        logger.error(f"Error buscando archivo base para tener la fecha base. key: {key}. {e}")
    return last_date


def save_df_to_s3(data: pd.DataFrame or str, conn: str, bucket: str, key: str, provider: str, delimiter: str = ",",
                  in_memory: bool = True):
    result = False
    try:
        logger.info(f"SUBIENDO A NUBE KEY {key}")
        file_type = get_type_file(key)
        gcp_cloud = False
        if provider == ProviderTypeEnum.AMAZON.value or provider == ProviderTypeEnum.MINIO.value:
            hook = S3Hook(conn)
        else:
            hook = GoogleCloudStorageHook(conn)
            gcp_cloud = True
        if file_type == FileTypeEnum.EXCEL or file_type == FileTypeEnum.OLD_EXCEL:
            if in_memory:
                with BytesIO() as buffer:
                    with pd.ExcelWriter(buffer, engine='xlsxwriter') as writer:
                        data.to_excel(writer, index=None)
                if gcp_cloud:
                    hook.upload(bucket, key, data=buffer.getvalue())
                else:
                    hook.load_bytes(buffer.getvalue(), key, bucket, True)
            else:
                pass
        elif file_type == FileTypeEnum.CSV or file_type == FileTypeEnum.TEXT:
            if in_memory:
                csv_buffer = BytesIO()
                data.to_csv(csv_buffer, header=True, index=False, sep=delimiter, na_rep='None')
                csv_buffer.seek(0)
                if gcp_cloud:
                    hook.upload(bucket, key, data=csv_buffer.getvalue())
                else:
                    hook.load_bytes(csv_buffer.getvalue(), key, bucket, True)
            else:
                if gcp_cloud:
                    hook.upload(bucket, key, data)
                else:
                    hook.load_file(data, key, bucket)
        result = True
    except Exception as e:
        logger.error(f"Error guardando archivos a S3. key: {key}. {e}")
    return result

def move_object_s3(conn: str, bucket: str, source_key: str, output_key: str) -> None:
    try:
        filename = source_key[source_key.rfind("/")+1:]
        output_key += filename
        s3_hook = S3Hook(conn)
        s3_hook.copy_object(source_key, output_key, bucket, bucket)
        s3_hook.delete_objects(bucket, source_key)
    except Exception as e:
        logger.error(f"Error moviendo archivo desde {source_key} hacia {output_key} en bucket {bucket}. {e}")


def get_files_from_prefix(conn: str, bucket: str, prefix: str, provider: str) -> List[Tuple[str, str]]:
    result = []
    allowed_filetypes = [ScriptFileTypeEnum[item].value for item in ScriptFileTypeEnum._member_names_]
    try:
        files = []
        s3_hook, gcp_hook, data = None, None, None
        if provider == ProviderTypeEnum.AMAZON.value or provider == ProviderTypeEnum.MINIO.value:
            s3_hook = S3Hook(conn)
            files = s3_hook.list_keys(bucket, prefix)
        elif provider == ProviderTypeEnum.GOOGLE.value:
            gcp_hook = GoogleCloudStorageHook(conn)
            if not prefix.endswith("/"):
                prefix += "/"
            files = gcp_hook.list(bucket, prefix=prefix)
        logger.debug(f"Archivos encontrados en el prefijo {prefix}: {files}")
        for file in files:
            if file.endswith("/") or file[file.rfind(".")+1:].lower() not in allowed_filetypes:
                continue
            if provider == ProviderTypeEnum.AMAZON.value or provider == ProviderTypeEnum.MINIO.value:
                data = s3_hook.get_key(file, bucket).get()['Body'].read().decode("utf-8")
            elif provider == ProviderTypeEnum.GOOGLE.value:
                data = gcp_hook.download(bucket, file).decode("utf-8")
            if file.find("/") == -1:
                filename = file
            else:
                filename = file[file.rfind("/")+1:]
            script = (filename, data)
            result.append(script)
    except Exception as e:
        logger.error(f"Error extrayendo archivos en memoria desde bucket {bucket} y prefix {prefix}. {e}")
    finally:
        return result


def get_file_from_prefix(conn: str, bucket: str, key: str, provider: str, timezone: str, task,
                         frequency: str = "montly") -> Any:
    result, key_result = BytesIO(), ''
    try:
        format_re_pattern = '[0-9]{4}-[0-9]{2}\.'
        format_date = "%Y-%m" if frequency == "montly" else "%Y-%W"
        period = str(datetime_by_tzone(timezone, format_date))[:7]
        logger.info(f"Periodo actual: {period}.")
        files, file = [], key
        cloud_gcp = False
        if provider == ProviderTypeEnum.AMAZON.value or provider == ProviderTypeEnum.MINIO.value:
            s3_hook = S3Hook(conn)
            files = s3_hook.list_keys(bucket, key)
        elif provider == ProviderTypeEnum.GOOGLE.value:
            cloud_gcp = True
            gcp_hook = GoogleCloudStorageHook(conn)
            if not key.endswith("/"):
                key += "/"
            files = gcp_hook.list(bucket, prefix=key)
        files_with_period = []
        for file in files:
            if file.endswith("/") or not re.search(format_re_pattern, file):
                continue
            file_period = file[file.rfind("_") + 1:file.rfind(".")]
            files_with_period.append((file, file_period))
        files_with_period.sort(key=lambda x: x[1])
        if len(files_with_period) > 0:
            file = files_with_period[-1][0]
            task.xcom_push(key="PROCESSED_PERIOD", value=files_with_period[-1][1])
        logger.info(f"Descargando archivo de control: {file}")
        if file != key:
            if cloud_gcp:
                result = gcp_hook.download(bucket, file)
                result = BytesIO(result)
            else:
                data = s3_hook.get_key(file, bucket)
                data.download_fileobj(result)
            key_result = file
    except Exception as e:
        result = None
        logger.error(f"Error extrayendo archivo {key}. {e}")
    finally:
        return result, key_result


def load_control_to_s3(obj, conn: str, bucket: str, key: str, provider: str, timezone: str,
                       frequency: str = "montly") -> bool:
    load = False
    try:
        format_date = "%Y-%m" if frequency == "montly" else "%Y-%W"
        period = str(datetime_by_tzone(timezone, format_date))[:7]
        key = key.replace("<period>", period)
        load = upload_file(obj, conn, bucket, key, provider)
    except Exception as e:
        logger.error(f"Error subiendo archivo de control a bucket {bucket} y key {key}. {e}")
    finally:
        return load


def load_report_to_s3(conn: str, bucket: str, key: str, filename: str, provider: str, timezone: str,
                      pattern: str) -> bool:
    load = False
    try:
        current_datetime = str(datetime_by_tzone(timezone, pattern))
        key = key.replace("<datetime>", current_datetime)
        if provider == ProviderTypeEnum.AMAZON.value or provider == ProviderTypeEnum.MINIO.value:
            s3_hook = S3Hook(conn)
            s3_hook.load_file(filename, key, bucket, True)
        elif provider == ProviderTypeEnum.GOOGLE.value:
            gcp_hook = GoogleCloudStorageHook(conn)
            gcp_hook.upload(bucket, key, filename)
        load = True
    except Exception as e:
        logger.error(f"Error subiendo reporte a bucket {bucket} y key {key}. {e}")
    finally:
        return load


def upload_file(obj, connection: str, bucket: str, key: str, provider: str) -> bool:
    upload = False
    try:
        if provider == ProviderTypeEnum.AMAZON.value or provider == ProviderTypeEnum.MINIO.value:
            s3_hook = S3Hook(connection)
            s3_hook.load_bytes(obj, key, bucket, True)
        elif provider == ProviderTypeEnum.GOOGLE.value:
            gcp_hook = GoogleCloudStorageHook(connection)
            gcp_hook.upload(bucket, key, data=obj)
        upload = True
    except Exception as e:
        logger.error(f"Error subiendo archivo a {provider}. bucket: {bucket}, key: {key}. {e}")
    finally:
        return upload
