from datetime import datetime
import time
from typing import Any, Dict
import json
from io import StringIO
from airflow import DAG
from airflow.operators.python import PythonOperator
from airflow.utils.task_group import TaskGroup
from components.Utils import update_sql_commands_2
from components.Xcom import save_commands_to_xcom
from components.S3Route import get_files_from_prefix, get_file_from_prefix
from components.Sensor import create_sensor
from components.Extractor import get_extract_task_group
from components.Transformation import get_transform_task_group
from components.Generation import get_generate_task_group
from components.Cleaning import get_cleaning_task_group
from enums.ProcessStatusEnum import ProcessStatusEnum

from components.Databases.Database import Database
import logging

logger = logging.getLogger()


DAG_NAME = "BCOM_DAG_EXTRACT_AND_TRANSFORM2"

# Change this path if is deployed in prod or dev
MAIN_PATH = "/opt/airflow/dags/"

DEFAULT_ARGS = {
    'owner': 'BCOM',
    "start_date": datetime(2023, 5, 25, 22, 9),
    'depends_on_past': False,
    'email': 'caguirre@bytesw.com',
    'retries': 0,
    'email_on_retry': False,
    'email_on_failure': False
}


def cleaning(intern_conn, control_s3: Dict[str, Any], provider: str, timezone: str) -> TaskGroup:
    groups = None
    try:
        groups = get_cleaning_task_group(intern_conn, control_s3, provider, timezone)
    except Exception as e:
        logger.error(f"Error general de transformación de datos. {e}")
    finally:
        return groups


def generate_and_deploy_results(intern_conn, parameters: Dict[str, Any], timezone: str,
                                control_s3: Dict[str, Any], provider: str) -> TaskGroup:
    groups = None
    try:
        groups = get_generate_task_group(intern_conn, parameters, control_s3, timezone, provider)
    except Exception as e:
        logger.error(f"Error general de creación y despliegue de resultados. {e}")
    finally:
        return groups


def transformation(intern_conn, timezone: str, control_s3: Dict[str, Any], provider: str, store_procedure: Dict, procedure_mask: str,
                   label_tablename: str, transform_mask: str, order_delimiter: str) -> TaskGroup:
    groups = None
    try:
        groups = get_transform_task_group(intern_conn, timezone, control_s3, provider, store_procedure, procedure_mask, label_tablename, transform_mask, order_delimiter)
    except Exception as e:
        logger.error(f"Error general de transformación de datos. {e}")
    finally:
        return groups


def extraction(source_conn, intern_conn, timezone: str, control_s3: Dict[str, Any],
               provider: str, chunksize: int = 100000) -> TaskGroup:
    groups = None
    try:
        groups = get_extract_task_group(source_conn, intern_conn, chunksize, timezone, control_s3, provider)
    except Exception as e:
        logger.error(f"Error general de extracción de datos. {e}")
    finally:
        return groups


def save_procedure_json(json_path: str, task) -> None:
    try:
        with open(json_path) as f:
            data = json.load(f)
        if data:
            task.xcom_push(key="EXTRACTION-DEFINITION-JSON", value=data)
    except Exception as e:
        logger.error(f"Error leyendo y guardando archivo descriptor de procedure. {e}")


def extract_control(conn_id: str, bucket: str, prefix: str, task, provider: str, timezone: str):
    try:
        if not prefix.endswith("/"):
            prefix += "/"
        logger.info(f"BUSCANDO Y EXTRAYENDO ARCHIVO DE CONTROL DESDE {prefix}")
        control, _ = get_file_from_prefix(conn_id, bucket, prefix, provider, timezone, task)
        logger.debug(f"Archivo de control descargado: {control}")
        try:
            str_data = str(control.getvalue(), encoding='UTF-8', errors='ignore')
            data = StringIO(str_data)
            data = json.load(data)
        except Exception:
            data = [{"status": ProcessStatusEnum.SUCCESS.value, "tasks": {}}]
            logger.info(f"Json de control creado: {data}")
        task.xcom_push(key="CONTROL-CONFIG", value=data)

    except Exception as e:
        logger.error(f"Error general de descarga de archivo de control. {e}")


def extract_scripts(conn_id: str, bucket: str, prefix: str, source_mask: str, transform_mask: str,
                    order_delimiter: str, procedure_mask: str, label_tablename: str, control_params: Dict[str, Any],
                    provider: str, timezone: str, procedure_filepath: str, **kwargs):
    try:
        extract_control(control_params["connection_id"], control_params["bucket"], control_params["prefix"],
                        kwargs['ti'], provider, timezone)
        save_procedure_json(procedure_filepath, kwargs['ti'])
        start_time = time.time()
        logger.info(f"EXTRAYENDO SCRIPTS DESDE {bucket}/{prefix}")
        scripts = get_files_from_prefix(conn_id, bucket, prefix, provider)
        scripts = update_sql_commands_2(scripts, label_tablename, source_mask)
        save_commands_to_xcom(scripts, kwargs['ti'], source_mask, transform_mask, procedure_mask, order_delimiter)
        logger.debug(f"Script cargados en Xcom: {scripts}")
        logger.info(f"Tiempo del Task de descarga de scripts: {round(time.time() - start_time, 3)} segundos")
    except Exception as e:
        raise AssertionError(f"Error general de descarga y guardado de scripts en variables Xcom. {e}")


def set_dag():
    """ DAG that execute a series of SQL commands from scripts to make a result file and save on a bucket (for now)"""
    import yaml
    from yaml.loader import SafeLoader

    # Cambiar conf_path dependiendo del ambiente, en prod usando k8 y contenedores usar /opt/airflow/dags/app_conf.yml
    # En desarrollo, cualquiera que apunte a su carpeta dags
    conf_path = MAIN_PATH + "dag_conf.yml"
    with open(conf_path) as f:
        data = yaml.load(f, Loader=SafeLoader)
        conf = data["app"]
    with DAG(DAG_NAME, default_args=DEFAULT_ARGS, description="Proceso que extrae y transforma",
             schedule_interval=conf["schedule"], tags=["DAG BCOM - SQL TRANSFORMATIONS"], catchup=False) as dag:
        scripts_s3 = conf["scripts"]["s3_params"]
        if scripts_s3["prefix"].endswith("/"):
            wildcard_scripts = scripts_s3["prefix"] + "?*"
        else:
            wildcard_scripts = scripts_s3["prefix"] + "/?*"
        sensor_scripts = create_sensor("SCRIPTS-SENSOR", scripts_s3["connection_id"], scripts_s3["bucket"],
                                       wildcard_scripts, conf["cloud_provider"])
        control_s3 = conf["control"]["s3_params"]
        timezone = conf["timezone"]
        # NUEVA VARIABLE
        procedure = conf["definitions"]
        # Scripts extraction
        extract_mask = conf["source_mask"]
        transform_mask = conf["transformation_mask"]
        procedure_mask = conf["procedure_mask"]
        order_delimiter = conf["prefix_order_delimiter"]
        script_extractor = PythonOperator(
            task_id="SCRIPTS-EXTRACTOR",
            python_callable=extract_scripts,
            op_kwargs={'conn_id': scripts_s3["connection_id"], 'bucket': scripts_s3["bucket"],
                       'prefix': scripts_s3["prefix"], 'source_mask': extract_mask, 'transform_mask': transform_mask,
                       'procedure_mask': procedure_mask, 'order_delimiter': order_delimiter,
                       'label_tablename': conf["label_multiple_select"], 'control_params': control_s3,
                       'provider': conf["cloud_provider"], 'timezone': timezone,
                       'procedure_filepath': procedure["filepath"]},
            trigger_rule="all_success"
        )

        # Source Database configuration. Only we use 1 source
        source_params = conf["database"]["sources"]["source1"]
        source_db = Database(source_params["type"], source_params["host"], int(source_params["port"]),
                             source_params["username"], source_params["password"], source_params["database"],
                             source_params["service"], source_params["schema"])
        source_db.create_engine()

        # Intern Database configuration.
        intern_params = conf["database"]["transformation"]
        intern_db = Database(intern_params["type"], intern_params["host"], int(intern_params["port"]),
                             intern_params["username"], intern_params["password"], intern_params["database"],
                             intern_params["service"], intern_params["schema"])
        intern_db.create_engine()

        # Creación de grupo de tasks para las extracciones
        chunksize = conf["chunksize"]
        extractions = extraction(source_db, intern_db, timezone, control_s3, conf["cloud_provider"], chunksize)

        # Creación de grupo de tasks para las transformaciones
        store_procedure = conf["store_procedures"]
        transformations = transformation(intern_db, timezone, control_s3, conf["cloud_provider"], store_procedure, procedure_mask, conf["label_transform_procedure"],
                                         transform_mask, order_delimiter)

        # Creación de grupo de tasks para la generación y despliegue de archivos resultados
        outputs_conf = conf["outputs"]

        result = generate_and_deploy_results(intern_db, outputs_conf, timezone, control_s3, conf["cloud_provider"])

        # Creación de tasks de limpiadores
        cleaners = cleaning(intern_db, control_s3, conf["cloud_provider"], timezone)

        sensor_scripts >> script_extractor >> extractions >> transformations >> result >> cleaners
    return dag


globals()["0"] = set_dag()

